import pathlib
import yamale

current_dir = (pathlib.Path(__file__).resolve().parent)
svs_config_schema_path = current_dir / 'svs-config-schema.yml'
svs_config_example_path = current_dir / 'svs-config-example.yml'

svs_config_schema = yamale.make_schema(svs_config_schema_path)


import numpy as np
import dask.array as da
import skimage.exposure
import sklearn.linear_model
from loguru import logger
import tqdm.dask

from . import register
from . import block_affine

class ReaderAligner:
    # aligner only uses single-channel form to register readers

    def __init__(self, ref_reader, moving_reader, pyramid_level=0):
        self.ref_reader = ref_reader
        self.moving_reader = moving_reader
        self.pyramid_level = pyramid_level

    @property
    def ref_img(self):
        return self.ref_reader.get_processed_color(
            level=self.pyramid_level, mode='grayscale'
        )

    @property
    def grid_shape(self):
        return self.ref_img.numblocks

    @property
    def num_blocks(self):
        return self.ref_img.npartitions
    
    def coarse_register_affine(self, n_keypoints=2000):
        logger.info(f"Coarse aligning to {self.ref_reader.path.name}")
        ref_img = self.ref_reader.get_processed_color(
            level=max(self.ref_reader.level_downsamples.keys())
        )
        moving_img = self.moving_reader.get_processed_color(
            level=max(self.moving_reader.level_downsamples.keys())
        )
        # `register.feature_based_registration` expects bright-field image to be
        # uint8 pixel dtype
        ref_img = skimage.exposure.rescale_intensity(
            ref_img, out_range=np.uint8
        )
        moving_img = skimage.exposure.rescale_intensity(
            moving_img, out_range=np.uint8
        )
        affine_matrix = register.feature_based_registration(
            ref_img, moving_img,
            n_keypoints=n_keypoints,
            plot_match_result=True
        )
        self.coarse_affine_matrix = np.vstack(
            [affine_matrix, [0, 0, 1]]
        )

    @property
    def affine_matrix(self):
        if not hasattr(self, 'coarse_affine_matrix'):
            self.coarse_register_affine()
        level_downsamples = self.moving_reader.level_downsamples
        coarse_factor = (
            level_downsamples[max(level_downsamples.keys())] /
            level_downsamples[self.pyramid_level]            
        )
        affine_matrix = self.coarse_affine_matrix.copy()
        affine_matrix[:2, 2] *= coarse_factor
        return affine_matrix

    @property
    def tform(self):
        return skimage.transform.AffineTransform(
            matrix=self.affine_matrix
        )

    def affine_transformed_moving_img(self, mode):
        ref_img = self.ref_img
        moving_img = self.moving_reader.get_processed_color(
            level=self.pyramid_level, mode=mode
        )
        return da.map_blocks(
            block_affine.block_affine_dask,
            self.affine_matrix,
            chunks=ref_img.chunks,
            dtype=ref_img.dtype,
            src_array=moving_img
        )

    def compute_shifts(self):
        logger.info(f"Computing block-wise shifts")
        ref_img = self.ref_img
        moving_img = self.affine_transformed_moving_img(mode='grayscale')
        shifts_da = da.map_blocks(
            lambda a, b: np.atleast_2d(
                register.phase_cross_correlation(a, b, sigma=0, upsample=1)[0]
            ),
            ref_img,
            moving_img,
            dtype=np.float32
        )
        with tqdm.dask.TqdmCallback(
            ascii=True, desc='Computing shifts',
        ):
            shifts = shifts_da.compute()
        self.shifts = shifts.reshape(-1, 2)

    def constrain_shifts(self):
        num_rows, num_cols = self.grid_shape
        self.distances = np.linalg.norm(self.shifts, axis=1)
        self.threshold_distance = skimage.filters.threshold_triangle(
            self.distances
        )

        high_confidence_blocks = self.distances < self.threshold_distance

        lr = sklearn.linear_model.LinearRegression()
        block_coords = np.mgrid[:num_rows, :num_cols].reshape(2, -1).T
        lr.fit(
            block_coords[high_confidence_blocks],
            self.shifts[high_confidence_blocks]
        )
        predicted_shifts = lr.predict(block_coords)
        diffs = self.shifts - predicted_shifts
        distance_diffs = np.linalg.norm(diffs, axis=1)
        passed = (
            distance_diffs <
            skimage.filters.threshold_triangle(distance_diffs)
        )
        self.shifts[~passed] = predicted_shifts[~passed]
    
    @property 
    def shifts_da(self):
        return da.from_array(
            self.shifts.reshape(self.grid_shape[0], -1),
            chunks=(1, 2)
        )
    
    @property
    def block_affine_matrices(self):
        mx = self.affine_matrix
        shifts = self.shifts
        
        def shift_affine_mx(mx, shift):
            y, x = shift
            mx_shift = np.eye(3)
            mx_shift[:2, 2] = x, y
            return mx @ mx_shift

        mxs = [
            shift_affine_mx(mx, s)
            for s in shifts
        ]
        return np.array(mxs)

    @property
    def block_affine_matrices_da(self):
        grid = np.arange(self.num_blocks).reshape(self.grid_shape)
        mxs = np.vstack([
            np.hstack(self.block_affine_matrices[r])
            for r in grid
        ])
        return da.from_array(mxs, chunks=3)
    
    def get_ref_mosaic(self, mode):
        level = self.pyramid_level
        ref_reader = self.ref_reader
        if mode == 'color':
            ref_img = da.moveaxis(
                ref_reader.pyramid_color[level], 2, 0
            )
        else:
            img_dtype = self.ref_reader.pixel_dtype
            _ref_img = ref_reader.get_processed_color(
                level=level, mode=mode
            )
            ref_img = da.map_blocks(
                skimage.exposure.rescale_intensity,
                _ref_img,
                chunks=_ref_img.chunks,
                dtype=img_dtype,
                in_range=(0, 1), 
                out_range=img_dtype.type
            )[np.newaxis, :]
        return ref_img

    def get_aligned_mosaic(self, mode):
        level = self.pyramid_level
        ref_reader = self.ref_reader
        moving_reader = self.moving_reader
        if mode == 'color':
            ref_img = da.moveaxis(
                ref_reader.pyramid_color[level], 2, 0
            )
            moving_img = da.moveaxis(
                moving_reader.pyramid_color[level], 2, 0
            )
            mxs = da.array([self.block_affine_matrices_da])
            multichannel = True
        else:
            ref_img = ref_reader.get_processed_color(
                level=level, mode=mode
            )
            moving_img = moving_reader.get_processed_color(
                level=level, mode=mode
            )
            mxs = self.block_affine_matrices_da
            multichannel = False
            img_dtype = self.ref_reader.pixel_dtype

        warped_img = da.map_blocks(
            block_affine.block_affine_dask,
            mxs,
            chunks=ref_img.chunks,
            dtype=ref_img.dtype,
            src_array=moving_img,
            multichannel=multichannel
        )

        if mode == 'color':
            return warped_img
        
        else:
            # output image in (C, Y, X) shape
            return da.map_blocks(
                skimage.exposure.rescale_intensity,
                warped_img,
                chunks=ref_img.chunks,
                dtype=img_dtype,
                in_range=(0, 1), 
                out_range=img_dtype.type
            )[np.newaxis, :]




import cv2
import numpy as np
import scipy.fft
from . import img_util

import skimage.feature
import skimage.registration
import itertools
import warnings
import logging

if hasattr(skimage.registration, 'phase_cross_correlation'):
    register_translation = skimage.registration.phase_cross_correlation
else:
    register_translation = skimage.feature.register_translation


# 
# Image-based registration
# 
def phase_cross_correlation(img1, img2, sigma, upsample=10, module='skimage'):
    assert module in ['cv2', 'skimage']
    
    img1w = img_util.whiten(img1, sigma)
    img2w = img_util.whiten(img2, sigma)
    
    if module == 'skimage':
        img1_f = scipy.fft.fft2(img1w)
        img2_f = scipy.fft.fft2(img2w)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', 'invalid value encountered in true_divide',
                RuntimeWarning,
            )
            shift, _error, _phasediff = register_translation(
                img1_f, img2_f, upsample_factor=upsample, space='fourier'
            )
    
    elif module == 'cv2':
        shift_xy, _response = cv2.phaseCorrelate(img1w, img2w)
        shift = shift_xy[::-1]

    # At this point we may have a shift in the wrong quadrant since the FFT
    # assumes the signal is periodic. We test all four possibilities and return
    # the shift that gives the highest direct correlation (sum of products).
    shape = np.array(img1.shape)
    shift_pos = (shift + shape) % shape
    shift_neg = shift_pos - shape
    shifts = list(itertools.product(*zip(shift_pos, shift_neg)))
    correlations = [
        np.abs(np.sum(img1w * cv2_translate(img2w, s)))
        for s in shifts
    ]
    idx = np.argmax(correlations)
    shift = shifts[idx]
    correlation = correlations[idx]
    total_amplitude = np.linalg.norm(img1w) * np.linalg.norm(img2w)
    if correlation > 0 and total_amplitude > 0:
        error = -np.log(correlation / total_amplitude)
    else:
        error = np.inf
    return shift, error


def cv2_translate(img, shift):
    assert img.ndim == len(shift) == 2
    sy, sx = shift
    return cv2.warpAffine(
        img,    
        np.array([[1, 0, sx], [0, 1, sy]], dtype=float),
        img.shape[::-1]
    )


def normalized_phase_correlation(img1, img2, sigma):
    w1 = img_util.whiten(img1, sigma)
    w2 = img_util.whiten(img2, sigma)
    corr = scipy.fft.fftshift(np.abs(scipy.fft.ifft2(
        scipy.fft.fft2(w1) * scipy.fft.fft2(w2).conj()
    )))
    corr /= (np.linalg.norm(w1) * np.linalg.norm(w2))
    return corr


# 
# Feature-based registration
# 
import skimage.filters
import matplotlib.pyplot as plt
import skimage.exposure


def feature_based_registration(
    img_left, img_right,
    n_keypoints=1000, plot_match_result=False,
    plot_individual_result=False, ransacReprojThreshold=5
):
    compare_funcs = [
        np.greater if i.max() > 255 else np.less
        for i in (img_left, img_right)
    ]
    imgs_otsu = [
        f(i, skimage.filters.threshold_otsu(i)).astype(np.uint8)
        for (i, f) in zip((img_left, img_right), compare_funcs)
    ]
    imgs_tri = [
        f(i, skimage.filters.threshold_triangle(i)).astype(np.uint8)
        for (i, f) in zip((img_left, img_right), compare_funcs)
    ]
    img_left, img_right = match_bf_fl_histogram(img_left, img_right)
    imgs_whiten = [
        img_util.whiten(i, 1)
        for i in (img_left, img_right)
    ]

    all_found = [
        cv2_feature_detect_and_match(
            *img_pair, n_keypoints=n_keypoints,
            plot_match_result=plot_individual_result
        )
        for img_pair in [
            imgs_otsu,
            imgs_tri,
            (img_left, img_right),
            imgs_whiten
        ]
    ]
    all_src = np.vstack([i[0] for i in all_found])
    all_dst = np.vstack([i[1] for i in all_found])

    t_matrix, mask = cv2.estimateAffine2D(
        all_dst, all_src, 
        method=cv2.RANSAC,
        ransacReprojThreshold=ransacReprojThreshold,
        maxIters=5000
    )
    if plot_match_result == True:
        plt.figure()
        plt.gray()
        skimage.feature.plot_matches(
            plt.gca(), img_left, img_right,
            np.fliplr(all_src), np.fliplr(all_dst),
            np.arange(len(all_src)).repeat(2).reshape(-1, 2)[mask.flatten()>0],
            only_matches=False
        )
    return t_matrix


def match_bf_fl_histogram(img1, img2):
    img_pair = (img1, img2)
    bf1, bf2 = [
        True if i.max() <= 255 else False 
        for i in img_pair
    ]
    if bf1 == bf2:
        return img1, img2
    elif bf1 is True:
        return img1, skimage.exposure.match_histograms(-img2, img1)
    elif bf2 is True:
        return skimage.exposure.match_histograms(-img1, img2), img2


def plot_img_keypoints(imgs, keypoints):
    fig, axs = plt.subplots(1, len(imgs))
    for i, k, a in zip(imgs, keypoints, axs):
        a.imshow(cv2.drawKeypoints(
            i, k, None,
            flags=cv2.DRAW_MATCHES_FLAGS_DEFAULT
        ))
        a.set_title(len(k))
    return 


def cv2_feature_detect_and_match(
    img_left, img_right, n_keypoints=1000,
    plot_keypoint_result=False, plot_match_result=False
):
    img_left, img_right = [
        img_util.cv2_to_uint8(i)
        for i in (img_left, img_right)
    ]
    descriptor_extractor = cv2.ORB_create(n_keypoints)

    keypoints_left, descriptors_left = descriptor_extractor.detectAndCompute(
        np.dstack(3*(img_left,)), None
    )
    keypoints_right, descriptors_right = descriptor_extractor.detectAndCompute(
        np.dstack(3*(img_right,)), None
    )
    if plot_keypoint_result == True:
        plot_img_keypoints(
            [img_left, img_right], [keypoints_left, keypoints_right]
        )
    logging.info(f"keypts L:{len(keypoints_left)}, keypts R:{len(keypoints_right)}")
    if len(keypoints_left) == 0 or len(keypoints_right) == 0:
        return np.empty((1, 2)), np.empty((1, 2))

    bf_matcher = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True)
    matches = bf_matcher.match(descriptors_left, descriptors_right)

    src_pts = np.float32(
        [keypoints_left[m.queryIdx].pt for m in matches]
    )
    dst_pts = np.float32(
        [keypoints_right[m.trainIdx].pt for m in matches]
    )
    t_matrix, mask = cv2.estimateAffine2D(
        dst_pts, src_pts, 
        method=cv2.RANSAC, ransacReprojThreshold=3, maxIters=5000
    )
    if plot_match_result == True:
        plt.figure()
        imgmatch_ransac = cv2.drawMatches(
            img_left, keypoints_left,
            img_right, keypoints_right,
            matches, None,
            matchColor=(0, 255, 0), singlePointColor=None,
            matchesMask=mask.flatten(),
            flags=cv2.DRAW_MATCHES_FLAGS_DEFAULT
        )
        plt.gca().imshow(imgmatch_ransac)
    return src_pts[mask.flatten()>0], dst_pts[mask.flatten()>0]









import numpy as np
import cv2


def cv2_pyramid(img, max_size=1024):
    size_max = np.array(img.shape).max()
    pyramid = [img]
    while size_max > max_size:
        img = cv2.pyrDown(img)
        pyramid.append(img)
        size_max //= 2
    return pyramid


def whiten(img, sigma=1):
    border_mode = cv2.BORDER_REFLECT
    g_img = img if sigma == 0 else cv2.GaussianBlur(
        img, (0, 0), sigma, 
        borderType=border_mode
    ).astype(np.float32)
    log_img = cv2.Laplacian(
        g_img, cv2.CV_32F, ksize=1,
        borderType=border_mode
    )
    # log_img = cv2.convertScaleAbs(log_img)
    return log_img


def cv2_to_uint8(img):
    return cv2.normalize(
        src=img, dst=None,
        alpha=255, beta=0,
        norm_type=cv2.NORM_MINMAX,
        dtype=cv2.CV_8U
    )


import math
import tifffile
import tqdm
import numpy as np
import skimage.transform
from loguru import logger

from . import __version__ as _version


class PyramidSetting:

    def __init__(
        self,
        downscale_factor=2,
        tile_size=1024,
        max_pyramid_img_size=1024
    ):
        self.downscale_factor = downscale_factor
        self.tile_size = tile_size
        self.max_pyramid_img_size = max_pyramid_img_size

    def tile_shapes(self, base_shape):
        shapes = np.array(self.pyramid_shapes(base_shape))
        n_rows_n_cols = np.ceil(shapes / self.tile_size)
        tile_shapes = np.ceil(shapes / n_rows_n_cols / 16) * 16
        return [tuple(map(int, s)) for s in tile_shapes]

    def pyramid_shapes(self, base_shape):
        num_levels = self.num_levels(base_shape)
        factors = self.downscale_factor ** np.arange(num_levels)
        shapes = np.ceil(np.array(base_shape) / factors[:,None])
        return [tuple(map(int, s)) for s in shapes]

    def num_levels(self, base_shape):
        factor = max(base_shape) / self.max_pyramid_img_size
        return math.ceil(math.log(factor, self.downscale_factor)) + 1


def format_channel_names(mosaics, channel_names):
    n_channels_each = [
        count_num_channels([m])
        for m in mosaics
    ]
    unique_names = make_unique_str(channel_names)
    channel_names = [
        [n]*c
        for c, n in zip(n_channels_each, unique_names)
    ]
    return [n for l in channel_names for n in l]


def make_unique_str(str_list):
    if len(set(str_list)) == len(str_list):
        return str_list
    else:
        max_length = max([len(s) for s in str_list])
        str_np = np.array(str_list, dtype=np.dtype(('U', max_length+10)))
        unique, counts = np.unique(str_np, return_counts=True)
        has_duplicate = unique[counts > 1]
        for n in has_duplicate:
            suffixes = [
                f"_{i}"
                for i in range(1, (str_np == n).sum()+1)
            ]
            str_np[str_np == n] = np.char.add(n, suffixes)
    return make_unique_str(list(str_np))


def write_pyramid(
    mosaics,
    output_path,
    pixel_size=1,
    channel_names=None,
    verbose=True,
):
    ref_m = mosaics[0]
    path = output_path
    num_channels = count_num_channels(mosaics)
    base_shape = ref_m.shape[1:3]
    downscale_factor = 4
    pyramid_setting = PyramidSetting(
        downscale_factor=downscale_factor,
        tile_size=max(ref_m.chunksize)
    )
    num_levels = pyramid_setting.num_levels(base_shape)
    tile_shapes = pyramid_setting.tile_shapes(base_shape)
    shapes = pyramid_setting.pyramid_shapes(base_shape)

    dtype = ref_m.dtype

    software = f'palom {_version}'
    pixel_size = pixel_size
    metadata = {
        'Creator': software,
        'Pixels': {
            'PhysicalSizeX': pixel_size, 'PhysicalSizeXUnit': '\u00b5m',
            'PhysicalSizeY': pixel_size, 'PhysicalSizeYUnit': '\u00b5m'
        },
    }

    if channel_names is not None:
        names = format_channel_names(mosaics, channel_names)
        if len(names) == num_channels:
            metadata.update({
                'Channel': {'Name': names},
            })

    logger.info(f"Writing to {path}")
    with tifffile.TiffWriter(path, bigtiff=True) as tif:
        tif.write(
            data=tile_from_combined_mosaics(
                mosaics, tile_shape=tile_shapes[0]
            ),
            metadata=metadata,
            software=software,
            shape=(num_channels, *shapes[0]),
            subifds=int(num_levels - 1),
            dtype=dtype,
            tile=tile_shapes[0]
        )
        logger.info('Generating pyramid')
        for level, (shape, tile_shape) in enumerate(
            zip(shapes[1:], tile_shapes[1:])
        ):
            if verbose:
                logger.info(f"    Level {level+1} ({shape[0]} x {shape[1]})")
            tif.write(
                data=tile_from_pyramid(
                    path,
                    num_channels,
                    tile_shape=tile_shape,
                    downscale_factor=downscale_factor,
                    level=level,
                ),
                shape=(num_channels, *shape),
                subfiletype=1,
                dtype=dtype,
                tile=tile_shape
            )


def count_num_channels(imgs):
    for img in imgs:
        assert img.ndim == 2 or img.ndim == 3
    return sum([
        1 if img.ndim == 2 else img.shape[0]
        for img in imgs
    ])
    

def tile_from_combined_mosaics(mosaics, tile_shape):
    num_rows, num_cols = mosaics[0].shape[1:3]
    h, w = tile_shape
    n = len(mosaics)
    for idx, m in enumerate(mosaics):
        # the performance is heavily degraded without pre-computing the mosaic
        # channel
        with tqdm.dask.TqdmCallback(
            ascii=True, desc=f'Assembling mosaic {idx+1:2}/{n:2}',
        ):
            m = m.compute()
        for c in m:
            for y in range(0, num_rows, h):
                for x in range(0, num_cols, w):
                    yield np.array(c[y:y+h, x:x+w])
                # yield m[y:y+h, x:x+w].copy().compute()


def tile_from_pyramid(
    path,
    num_channels,
    tile_shape,
    downscale_factor=2,
    level=0
):
    h, w = tile_shape
    for c in tqdm.trange(
            num_channels,
            ascii=True, desc=f'Processing channel'
        ):
        img = tifffile.imread(
            path, is_ome=False, series=0, key=c, level=level
        )
        img = skimage.transform.downscale_local_mean(
            img, (downscale_factor, downscale_factor)
        ).astype(img.dtype)
        num_rows, num_columns = img.shape
        for y in range(0, num_rows, h):
            for x in range(0, num_columns, w):
                yield np.array(img[y:y+h, x:x+w])



import sys
import argparse
import pathlib
import yamale
from loguru import logger
import datetime
import shutil

from . import schema
from .. import reader, align, write_pyramid
from .. import __version__ as _version

import matplotlib.pyplot as plt


def main(argv=sys.argv):

    parser = argparse.ArgumentParser(
        description=(
            'Align multiple SVS images of the same biospecimen and writes a merged'
            ' pyramidal ome-tiff'
        )
    )
    parser.add_argument(
        '--version', dest='version', default=False, action='store_true',
        help='print version'
    )

    subparsers = parser.add_subparsers(
        title='subcommands',
        dest='subparser_name',
        help=(
            'show configuration yaml example and schema; run using a configuration file'
        )
    )

    parser_show = subparsers.add_parser('show')
    parser_show.add_argument(
        'info-type', choices=['example', 'schema'],
        help='show configuration yaml example/schema in console'
    )
    parser_show.set_defaults(func=show_info)

    parser_run = subparsers.add_parser(
        'run',
        description='process files described in a configuration yaml file'
    )
    parser_run.add_argument(
        '-c', '--config-file',
        metavar='FILE',
        type=argparse.FileType('r'),
        help=(
            'a configuration yaml file; see an example in console by running'
            ' `palom-svs show example`'
        ),
        required=True
    )
    parser_run.set_defaults(func=run)

    args = parser.parse_args(argv[1:])
    
    if len(argv) == 1:
        parser.print_help()
        return 0
    
    if args.version:
        print(f"palom v{_version}")
        return 0

    return args.func(args)


def show_info(args):
    info_type = vars(args)['info-type']

    if info_type == 'example':
        info_path = schema.svs_config_example_path
    else:
        info_path = schema.svs_config_schema_path
    with open(info_path) as f:
        print_text = f.read()
    print(f"\n{info_path}\n\n\n{print_text}\n\n")
    return 0


def run(args):
    config_file = args.config_file
    config_filepath = pathlib.Path(config_file.name)

    time_str = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    log_path = (
        config_filepath.parent /
        f"{config_filepath.name}-run-{time_str}.log"
    )
    logger_id = logger.add(log_path)

    logger.info(f"Running palom v{_version}")
    logger.info(f"Reading {config_file.name}\n\n{config_file.read().strip()}\n")
    config_file.close()

    config_data = yamale.make_data(config_file.name)
    if validate_config(config_data) == 1:
        return 1
    
    config = config_data[0][0]

    LEVEL = 0
    if 'pyramid level' in config:
        LEVEL = config['pyramid level']

    if 'pixel size' not in config:
        logger.warning(f"Pixel size in the output file is not set, using 1 µm")
        pixel_size = 1
    else: pixel_size = config['pixel size']

    images = get_image_list(config)
    
    image_paths = [
        pathlib.Path(config['input dir']) / pathlib.Path(i['filename'])
        for i in images
    ]

    image_output_modes = [
        i['output mode']
        for i in images
    ]

    channel_names = [
        i['channel name'] if 'channel name' in i else f"Channel {idx+1}"
        for idx, i in enumerate(images)
    ]

    output_path, qc_path = validate_output_path(config['output full path'])

    run_result = run_palom(
        img_paths=image_paths,
        img_modes=image_output_modes,
        pixel_size=pixel_size,
        channel_names=channel_names,
        output_path=output_path,
        qc_path=qc_path,
        level=LEVEL
    )

    logger.info(f"Finishing {config_file.name}")    
    logger.remove(logger_id)
    # Can't `pathlib.Path.replace` across different disk drive
    shutil.copy(log_path, qc_path / f"{output_path.name}.log")
    log_path.unlink()

    return run_result


def run_palom(
    img_paths,
    img_modes,
    pixel_size,
    channel_names,
    output_path,
    qc_path,
    level
):
    ref_reader = reader.SvsReader(img_paths[0])

    aligners = []
    for idx, p in enumerate(img_paths[1:]):
        logger.info(f"Processing {p.name}")
        moving_reader = reader.SvsReader(p)

        aligner = align.ReaderAligner(ref_reader, moving_reader, pyramid_level=level)
        aligner.coarse_register_affine()
        plt.suptitle(f"L: {aligner.ref_reader.path.name}\nR: {p.name}")
        plt.savefig(qc_path / f"{idx+1:02d}-{p.name}.png")
        aligner.compute_shifts()
        aligner.constrain_shifts()

        aligners.append(aligner)
    
    m1 = align.ReaderAligner(
        ref_reader, ref_reader, pyramid_level=level
    ).get_ref_mosaic(mode=img_modes[0])
    mosaics = [m1]
    mosaics += [
        a.get_aligned_mosaic(mode=m)
        for a, m in zip(aligners, img_modes[1:])
    ]

    write_pyramid.write_pyramid(
        mosaics, output_path,
        pixel_size=pixel_size,
        channel_names=channel_names
    )
    return 0


@logger.catch
def validate_output_path(out_path, overwrite=True):
    # write access
    out_img_path = pathlib.Path(out_path)
    if out_img_path.exists():
        logger.warning(f"{out_img_path} already exists and will be overwritten")
    out_qc_path = out_img_path.parent / 'qc'
    out_qc_path.mkdir(exist_ok=True, parents=True)
    # folder is not empty
    if any(out_qc_path.iterdir()):
        logger.warning(f"{out_qc_path} already exists and will be overwritten")
    return out_img_path, out_qc_path


def validate_config(config_data):
    try:
        yamale.validate(schema.svs_config_schema, config_data)
        logger.info('Config YAML validation success')
        return 0
    except yamale.YamaleError as e:
        logger.info('Config YAML validation failed')
        for result in e.results:
            logger.error(f"Error validating data '{result.data}' with '{result.schema}'")
            for error in result.errors:
                logger.error(f'\t{error}')
        return 1


def get_image_list(config):
    out_list = [config['reference image']]
    if 'moving images' not in config:
        return out_list
    else:
        if type(config['moving images']) == list:
            return out_list + config['moving images']
        else:
            return out_list + [config['moving images']]


if __name__ == '__main__':
    sys.exit(main())


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['palom', 'palom.cli', 'palom.cli.schema']

package_data = \
{'': ['*']}

install_requires = \
['dask>=2021.10.0,<2022.0.0',
 'loguru>=0.5.3,<0.6.0',
 'matplotlib>=3.4.3,<4.0.0',
 'napari-lazy-openslide>=0.2.0,<0.3.0',
 'numpy>=1.21.3,<2.0.0',
 'opencv-python>=4.5.3.56,<5.0.0.0',
 'scikit-image>=0.18.3,<0.19.0',
 'scikit-learn>=0.24.2,<0.25.0',
 'scipy>=1.7.1,<2.0.0',
 'tifffile>=2021.10.12,<2022.0.0',
 'tqdm>=4.62.3,<5.0.0',
 'yamale>=4.0.2,<5.0.0',
 'zarr>=2.10.0,<3.0.0']

entry_points = \
{'console_scripts': ['palom-svs = palom.cli.svs:main']}

setup_kwargs = {
    'name': 'palom',
    'version': '2021.11.1',
    'description': 'Piecewise alignment for layers of mosaics',
    'long_description': '# Piecewise alignment for layers of mosaics',
    'author': 'Yu-An Chen',
    'author_email': 'atwood12@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.8,<3.8',
}


setup(**setup_kwargs)



import pathlib
from napari_lazy_openslide import OpenSlideStore
import zarr
import dask.array as da
import numpy as np
import skimage.color
import skimage.exposure
from . import extract


class SvsReader:
    
    def __init__(self, path) -> None:
        self.path = pathlib.Path(path)
        self.store = OpenSlideStore(str(self.path))
        self.zarr = zarr.open(self.store, mode='r')

    @property
    def pixel_dtype(self):
        return self.pyramid_color[0].dtype

    @property
    def level_downsamples(self):
        return {
            i: int(np.round(d))
            for i, d in enumerate(self.store._slide.level_downsamples)
        }

    @property
    def pyramid_color(self):
        return [da.from_zarr(self.store, component=d['path'])[..., :3]
            for d in
            self.zarr.attrs['multiscales'][0]['datasets']
        ]

    def find_processed_color_contrast_range(self, mode):
        assert mode in ['grayscale', 'hematoxylin', 'intensity']
        
        mode_range = f'_{mode}_range'
        if hasattr(self, mode_range):
            return getattr(self, mode_range)
        
        thumbnail = self.pyramid_color[-1].compute()
        if mode == 'grayscale':
            img = self.rgb2gray(thumbnail)
        elif mode == 'hematoxylin':
            img = self.rgb2hematoxylin(thumbnail)
        elif mode == 'intensity':
            img = self.rgb2aec(thumbnail)

        setattr(self, mode_range, (
            img.min(), img.max()
        ))
        return getattr(self, mode_range)

    def rgb2gray(self, rgb_img):
        import cv2
        return cv2.cvtColor(rgb_img, cv2.COLOR_BGR2GRAY)
        # return skimage.color.rgb2gray(rgb_img).astype(np.float32)

    def rgb2aec(self, rgb_img):
        return extract.rgb2aec(rgb_img).astype(np.float32)

    def rgb2hematoxylin(self, rgb_img):
        hax = skimage.color.separate_stains(rgb_img, skimage.color.hax_from_rgb)
        return hax[..., 0].astype(np.float32)

    def get_processed_color(self, level, mode='grayscale'):
        assert mode in ['grayscale', 'hematoxylin', 'intensity']
        target = self.pyramid_color[level]

        if mode == 'grayscale':
            process_func = self.rgb2gray
        elif mode == 'hematoxylin':
            process_func = self.rgb2hematoxylin
        elif mode == 'intensity':
            process_func = self.rgb2aec

        intensity_range = self.find_processed_color_contrast_range(mode)
        
        processed = da.map_blocks(
            process_func,
            target,
            dtype=np.float32,
            drop_axis=2
        )

        return da.map_blocks(
            lambda x: skimage.exposure.rescale_intensity(
                x, in_range=intensity_range, out_range=(0, 1)
            ).astype(np.float32),
            processed,
            dtype=np.float32
        )



try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__version__ = importlib_metadata.version(__name__)


from . import (
    reader,
    align,
    write_pyramid,
    
    # debugging
    block_affine,
    extract,
    img_util,
    register
)


import skimage.util
import skimage.filters
import skimage.exposure
import numpy as np
import warnings

def imagej_rgb2cmyk(rgb_img):
    shape = rgb_img.shape
    assert 3 in shape, (
        'Image of shape {} is not an RGB image'.format(shape)
    )
    channel_idx = shape.index(3)
    rgb_img = skimage.util.img_as_float(rgb_img)
    rgb = np.moveaxis(rgb_img, channel_idx, 0)

    cmy = 1-rgb
    k = cmy.min(axis=0)
    
    s = 1-k

    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore', 'invalid value encountered in true_divide', RuntimeWarning,
        )
        f_cmy = (cmy-k) / s
    f_cmy *= ~(k >= 1)

    return np.append(f_cmy, k.reshape(1, *k.shape), axis=0)

def cmyk2marker_int(cmyk_img, min, max):
    marker_int = cmyk_img[1] + cmyk_img[2]
    marker_int = skimage.filters.median(marker_int, selem=np.ones((3,3)))

    marker_int = skimage.exposure.rescale_intensity(
        marker_int, in_range=(min, max), out_range=(0, 1)
    )
    return marker_int

def ohsu_cmyk2marker_int(cmyk_img):
    marker_int = cmyk_img[1] + cmyk_img[2]
    # the latest workflow applys median filter here while earlier
    # workflow seems to apply median filter after rescale_intensity
    marker_int = skimage.filters.median(marker_int, selem=np.ones((3,3)))

    max_int = marker_int.max()
    marker_int = skimage.exposure.rescale_intensity(
        marker_int.astype(float),
        in_range=(0.05*max_int, 0.95*max_int),
        out_range=(0, 1)
    )
    return skimage.util.img_as_ubyte(marker_int)

def rgb2aec(rgb_img):
    cmyk_img = imagej_rgb2cmyk(rgb_img)
    aec = cmyk_img[1] + cmyk_img[2]
    return skimage.filters.median(aec, selem=np.ones((3,3)))


import numpy as np
import skimage.transform

import cv2
import itertools


def block_affine_dask(
    affine_matrix,
    src_array=None,
    fill_empty=0,
    multichannel=False,
    block_info=None,
):
    if not multichannel:
        mx = affine_matrix
        Y, X = block_info[None]['array-location']
    else:
        mx = affine_matrix[0]
        C, Y, X = block_info[None]['array-location']
    transformation = skimage.transform.AffineTransform(matrix=mx)
    shape = block_info[None]['chunk-shape']
    return block_affine(
        (Y[0], X[0]), shape,
        transformation, src_array,
        fill_empty=fill_empty, multichannel=multichannel
    )


def block_affine(
    position, block_shape, 
    transformation, src_img,
    fill_empty=0, multichannel=False
):
    assert np.min(block_shape) >= 0, (
        f'block_shape {block_shape} is invalid' 
    )
    if multichannel:
        assert np.min(block_shape) == block_shape[0], (
            'multichannel block must has shape of (C, Y, X)'
        )
        assert np.min(src_img.shape) == src_img.shape[0], (
            'multichannel image must be shaped as (C, Y, X)'
        )
        channel, height, width = block_shape
        block_shape = (height, width, channel)
    else:
        assert len(src_img.shape) == 2, (
            'src_img has more than 2 dimensions, use '
            'multichannel=True for color image'
        )
        height, width = block_shape
    y0_dst, x0_dst = position
    y1_dst, x1_dst = ((height, width) + np.array(position)).astype(int)

    inversed_corners = transformation.inverse(
        list(itertools.product(
            [x0_dst, x1_dst], [y0_dst, y1_dst]
        ))
    )

    x1_src, y1_src = np.ceil(
        np.clip(inversed_corners.max(axis=0), 0, None)
    ).astype(int)
    x0_src, y0_src = np.floor(
        np.clip(inversed_corners.min(axis=0), 0, None)
    ).astype(int)
    
    if multichannel:
        src_img_block = src_img[:, y0_src:y1_src, x0_src:x1_src]
        src_img_block = np.moveaxis(src_img_block, 0, 2)
    else:
        src_img_block = src_img[y0_src:y1_src, x0_src:x1_src]
    src_dtype = src_img.dtype
    if 0 in src_img_block.shape:
        return np.ones(block_shape, dtype=src_dtype) * fill_empty
    translation_offset = transformation([[x0_src, y0_src]]) - (x0_dst, y0_dst)

    block_tform = skimage.transform.AffineTransform(
        translation=translation_offset,
        scale=transformation.scale,
        rotation=transformation.rotation,
        shear=transformation.shear
    )
    warped_src_img_block = cv2.warpAffine(
        src_img_block, block_tform.params[:2, :],
        (width, height)
    )
    if multichannel:
        # shape multichannel image as (C, Y, X)
        warped_src_img_block = np.moveaxis(warped_src_img_block, 2, 0)

    return warped_src_img_block
